import numpy
import sys
from getopt import getopt
from util import *
from data import *
from pylab import *
import re
from math import sqrt, ceil, floor
import os
from numpy.random import *
from gpumodel import GPUModel
import subprocess

MAX_FILTERS_PER_PIC = 64

cmReds = matplotlib.colors.LinearSegmentedColormap('reds', {'red':   [(0.0,  0.0, 0.0),
                                                                      (1.0,  1.0, 1.0)],
                                                            'green': [(0.0,  0.0, 0.0),
                                                                      (1.0,  0.0, 0.0)],
                                                            'blue':  [(0.0,  0.0, 0.0),
                                                                      (1.0,  0.0, 0.0)]})

cmGreens = matplotlib.colors.LinearSegmentedColormap('greens', {'green':  [(0.0,  0.0, 0.0),
                                                                           (1.0,  1.0, 1.0)],
                                                                'red':    [(0.0,  0.0, 0.0),
                                                                           (1.0,  0.0, 0.0)],
                                                                'blue':   [(0.0,  0.0, 0.0),
                                                                           (1.0,  0.0, 0.0)]})

cmBlues = matplotlib.colors.LinearSegmentedColormap('blues', {'blue':   [(0.0,  0.0, 0.0),
                                                                         (1.0,  1.0, 1.0)],
                                                              'green':  [(0.0,  0.0, 0.0),
                                                                         (1.0,  0.0, 0.0)],
                                                              'red':    [(0.0,  0.0, 0.0),
                                                                         (1.0,  0.0, 0.0)]})
def getpic(picvec):
    num_pixels = picvec.shape[0] / 3 # number of pixels in the image
    img_size = sqrt(num_pixels) # width (and also height) of the image
    y = picvec.reshape(3,img_size,img_size)
    y = y.swapaxes(0,2).swapaxes(0,1)
    return y

def draw_sub_fig(fignum, pic, filters_per_dim, subplot_number, filter_number, cmap=cm.gray, rescale=False, number_filters=True):
    if rescale:
        pic -= pic.min()
        pic /= pic.max()
    figure(fignum)
    subplot(filters_per_dim, filters_per_dim, subplot_number)
    draw_fig(fignum, pic, fig_title=number_filters and str(filter_number) or "", cmap=cmap) 
    xticks([])
    yticks([])
    
def draw_fig(fignum, pic, fig_title=None, cmap=None):
    figure(fignum)
    xticks([])
    yticks([])
    if not cmap is None:
        norm = Normalize(vmin=0, vmax=1)
        imshow(pic, interpolation='nearest', cmap=cmap, norm=norm)
    else:
        imshow(pic, interpolation='nearest')
    if fig_title:
        title(fig_title)
        
def print_usage():
    print "-f <rbm file>"
    print "-e show error plot"
    print "-l <N[-M]> - show these filters (hidden units)"
    print "\t-b - show filter color channels in separate figures"
    print "\t-r - rescale each filter individually"
    print "\t-n - number the filters"
    print "-g <N[-M]> - show these global filters (hidden units)"
    print "-s <dir> -- save all filter history as png to given dir (turns off all other options)"
    print "-h show histogram of weight distribution"

    #print "-i show histogram of weight increments"

def weights_hist(fignum, _title, weights):
    figure(fignum)
    title(_title)
    (n, bins, patches) = hist(weights, bins=100, fc='orange',linewidth=0)

def trim_filter_range(filter_range):
    m = re.match('^(\d+)(?:\-(\d+))?$', filter_range)
    r1 = int(m.group(1))
    r2 = m.group(2) and int(m.group(2)) or (r1 + MAX_FILTERS_PER_PIC - 1)
    filter_range = (r1, r2)
    if filter_range[1] - filter_range[0] + 1 > MAX_FILTERS_PER_PIC:
        print "Number of filters in range exceeds maximum of %d." % MAX_FILTERS_PER_PIC
        filter_range[1] = filter_range[0] + MAX_FILTERS_PER_PIC - 1
        print "Only filters %d to %d will be shown." % filter_range
    return filter_range

def plot_filters(hid_vis, filter_range, fignum, color_chan, rescale=False, number_filters=False, trans=True):
    hid_vis -= hid_vis.min()
    hid_vis /= hid_vis.max()
    if not trans:
        hid_vis = hid_vis.T
    color_chan_names = ("all colors", "reds", "greens", "blues")
    color_maps = (None, cmReds, cmGreens, cmBlues)
    filters_per_dim = int(ceil(sqrt(filter_range[1] - filter_range[0] + 1)))
    f = figure(fignum); clf()
    max_filter = min(filter_range[1], hid_vis.shape[0]-1)
    f.text(.5, .95, 'Filters %d to %d, %s' % (filter_range[0],max_filter, color_chan_names[color_chan]), horizontalalignment='center') 

    for y in xrange(0,filters_per_dim):
        for x in xrange(0,filters_per_dim):
            subplot_number = y * filters_per_dim + x + 1
            filter_number = filter_range[0] + subplot_number - 1
            if filter_number > min(filter_range[1], hid_vis.shape[0]-1):
                break
            filter = hid_vis[filter_number,:]

            filter_pic= getpic(filter)
            if color_chan == 0:
                draw_sub_fig(fignum, filter_pic, filters_per_dim, subplot_number, filter_number, rescale=rescale, number_filters=number_filters)
            else:
                draw_sub_fig(fignum, filter_pic[:,:,color_chan-1], filters_per_dim, subplot_number, filter_number, cmap=color_maps[color_chan], rescale=rescale, number_filters=number_filters)

if __name__ == "__main__":
    if len(sys.argv) == 1:
        print_usage()
        sys.exit(0)

    (options, args) = getopt(sys.argv[1:], "f:l:hebrns:g:")
    options = dict(options)
    rbm_file = options["-f"]
    filter_range = "-l" in options.keys() and options["-l"] or None
    filter_range_glob = "-g" in options.keys() and options["-g"] or None
    show_weights_hist = "-h" in options.keys()
    show_err = "-e" in options.keys()
    show_bw_filters = "-b" in options
    rescale = "-r" in options
    number_filters = "-n" in options
    filter_sequence_save_dir = "-s" in options and options["-s"] or None
    
    dic = GPUModel.load_checkpoint(rbm_file)
    state_dic = dic["model_state"]
    op = dic["op"].options
    dic["op"].print_values()
    #show_weights_inc_hist = "-i" in options.keys()
    if filter_range:
        filter_range = trim_filter_range(filter_range)
            
    num_hid = state_dic["hid_vis"].shape[0]
    numbatches = len(op["train_batch_range"].value)
    if show_err and not filter_sequence_save_dir:
        # plot test errors and train errors
        train_errors = [t[0] for t in state_dic["train_outputs"]]
        test_errors = numpy.row_stack([t[0] for t in state_dic["test_outputs"]])
        test_errors = numpy.tile(test_errors, (1, op["testing_freq"].value))
        test_errors = list(test_errors.flatten())
        test_errors += [test_errors[-1]] * (len(train_errors) - len(test_errors))

        numepochs = len(train_errors) / float(numbatches)
        figure(1)
        x = range(0, len(train_errors))
        plot(x, train_errors, 'k-', label='Training set')
        plot(x, test_errors, 'r-', label='Test set')
        legend()
        ticklocs = range(numbatches, len(train_errors) - len(train_errors) % numbatches + 1, numbatches)
        epoch_label_gran = int(ceil(numepochs / 20.))
        #ticklabels = map(lambda x: str((x / numbatches)), ticklocs)
        ticklabels = map(lambda x: x[0] % epoch_label_gran == 0 and str((x[1] / numbatches)) or '', enumerate(ticklocs))

        xticks(ticklocs, ticklabels)
        xlabel('Epoch')
        ylabel('Squared reconstruction error')
        title('Squared reconstruction error, h=%d, nb=%d' % (num_hid, numbatches))

    # plot some filters
    if filter_range:
        if filter_sequence_save_dir:
            rbm_file = rbm_file.rstrip("/")
            filter_sequence_save_dir = os.path.join(filter_sequence_save_dir, os.path.basename(rbm_file))
            files = sorted(os.listdir(rbm_file), key=alphanum_key)
            for i, f in enumerate(files):
                png_files = [os.path.join(filter_sequence_save_dir, x, "%s-%s.png" % (f, x)) for x in ("all", "reds", "greens", "blues")]
                if not os.path.exists(png_files[0]):
                    print "Saving filter checkpoint %d of %d ..." % (i + 1, len(files))
                    hid_vis = unpickle(os.path.join(rbm_file, f))["model_state"]["hid_vis"]
                    for i, out_file in enumerate(png_files):
                        if i > 0 and not show_bw_filters: break
                        if not os.path.exists(os.path.dirname(out_file)):
                            os.makedirs(os.path.dirname(out_file))
                        plot_filters(hid_vis, filter_range, i + 3, i, rescale=rescale, number_filters=number_filters)
                        savefig(out_file)
                        cmd = ['mogrify', '-trim', out_file]
                        subprocess.check_call(cmd)
            print "Saved filter checkpoint images to %s" % filter_sequence_save_dir
        else:
            plot_filters(state_dic["hid_vis"], filter_range, 2, 0, rescale=rescale, number_filters=number_filters)
            if show_bw_filters:
                plot_filters(state_dic["hid_vis"], filter_range, 4, 1, rescale=rescale, number_filters=number_filters)
                plot_filters(state_dic["hid_vis"], filter_range, 5, 2, rescale=rescale, number_filters=number_filters)
                plot_filters(state_dic["hid_vis"], filter_range, 6, 3, rescale=rescale, number_filters=number_filters)
            
            # show visible biases
            visb = state_dic["vis_biases"][0,:].copy()
            visb -= visb.min()
            visb /= visb.max()
            visb_pic = getpic(visb)
            draw_fig(70, visb_pic,fig_title="Visible biases")
            if show_bw_filters:
                draw_fig(71, visb_pic[:,:,0],fig_title="Visible biases (reds)", cmap=cmReds)
                draw_fig(72, visb_pic[:,:,1],fig_title="Visible biases (greens)", cmap=cmGreens)
                draw_fig(73, visb_pic[:,:,2],fig_title="Visible biases (blues)", cmap=cmBlues)
    
    if filter_range_glob:
        plot_filters(state_dic["vis_hid_glob"], filter_range, 40, 0, rescale=rescale, number_filters=number_filters, trans=False)
        if show_bw_filters:
            plot_filters(state_dic["vis_hid_glob"], filter_range, 41, 1, rescale=rescale, number_filters=number_filters, trans=False)
            plot_filters(state_dic["vis_hid_glob"], filter_range, 51, 2, rescale=rescale, number_filters=number_filters, trans=False)
            plot_filters(state_dic["vis_hid_glob"], filter_range, 61, 3, rescale=rescale, number_filters=number_filters, trans=False)
            
    if show_weights_hist and not filter_sequence_save_dir:
        hid_vis = state_dic["hid_vis"]
        hid_vis_inc = state_dic["hid_vis_inc"]
        vis_biases = state_dic["vis_biases"]
        hid_biases = state_dic["hid_biases"]
        vis_bias_inc = state_dic["vis_bias_inc"]
        hid_bias_inc = state_dic["hid_bias_inc"]
        hid_means = state_dic["hid_means"]

        weights_hist(60, 'Histogram of pixel-conv weights, h=%d' % (num_hid), hid_vis.flatten())

        # show histogram of weight increments
        #print vis_hid_inc
        #print (vis_hid_inc.min(), vis_hid_inc.max())
        weights_hist(7, 'Histogram of pixel-conv weight increments',hid_vis_inc.flatten())

        # show histogram of visible biases
        weights_hist(8, 'Histogram of vis biases',vis_biases.flatten())
        # show histogram of hid biases
        weights_hist(9, 'Histogram of hidden biases',hid_biases.flatten())

        # show histogram of visible bias increments
        weights_hist(10, 'Histogram of vis bias increments',vis_bias_inc.flatten())
        # show histogram of hid bias increments
        weights_hist(11, 'Histogram of hidden bias increments',hid_bias_inc.flatten())
        
        weights_hist(12, 'Histogram of hidden means',hid_means.flatten())

    if not filter_sequence_save_dir:
        show()

